from urllib.parse import urlencode
from ebrains_drive.files import SeafDir, SeafFile
from ebrains_drive.utils import raise_does_not_exist


class Repo(object):
    """
    A seafile library
    """

    def __init__(self, client, **kwargs):
        self.client = client

        allowed_keys = [
            "encrypted",
            "group_name",
            "groupid",
            "head_commit_id",
            "id",
            "modifier_contact_email",
            "modifier_email",
            "modifier_name",
            "mtime",
            "mtime_relative",
            "name",
            "owner",
            "owner_contact_email",
            "owner_name",
            "permission",
            "root",
            "share_from",
            "share_from_contact_email",
            "share_from_name",
            "share_type",
            "size",
            "size_formatted",
            "type",
            "version",
            "virtual",
        ]
        # Update __dict__ but only for keys that have been predefined
        # (silently ignore others)
        self.__dict__.update((key, value) for key, value in kwargs.items() if key in allowed_keys)
        # To NOT silently ignore rejected keys
        # rejected_keys = set(kwargs.keys()) - set(allowed_keys)
        # if rejected_keys:
        #     raise ValueError("Invalid arguments in constructor:{}".format(rejected_keys))

    def __str__(self):
        return "(id='{}', name='{}')".format(self.id, self.name)

    def __repr__(self):
        return "ebrains_drive.repo.Repo(id='{}', name='{}')".format(self.id, self.name)

    @classmethod
    def from_json(cls, client, repo_json):
        return cls(client, **repo_json)

    def is_readonly(self):
        return "w" not in self.perm

    @raise_does_not_exist("The requested file does not exist")
    def get_file(self, path):
        """Get the file object located in `path` in this repo.

        Return a :class:`SeafFile` object
        """
        assert path.startswith("/")
        url = "/api2/repos/%s/file/detail/" % self.id
        query = "?" + urlencode(dict(p=path))
        file_json = self.client.get(url + query).json()

        return SeafFile(self, path, file_json["id"], "file", file_json["size"])

    @raise_does_not_exist("The requested dir does not exist")
    def get_dir(self, path):
        """Get the dir object located in `path` in this repo.

        Return a :class:`SeafDir` object
        """
        assert path.startswith("/")
        url = "/api2/repos/%s/dir/" % self.id
        query = "?" + urlencode(dict(p=path))
        resp = self.client.get(url + query)
        dir_id = resp.headers["oid"]
        dir_json = resp.json()
        dir = SeafDir(self, path, dir_id, "dir")
        dir.load_entries(dir_json)
        return dir

    def delete(self):
        """Remove this repo. Only the repo owner can do this"""
        self.client.delete("/api2/repos/" + self.id)

    def list_history(self):
        """List the history of this repo

        Returns a list of :class:`RepoRevision` object.
        """
        pass

    ## Operations only the repo owner can do:

    def update(self, name=None):
        """Update the name of this repo. Only the repo owner can do
        this.
        """
        pass

    def get_settings(self):
        """Get the settings of this repo. Returns a dict containing the following
        keys:

        `history_limit`: How many days of repo history to keep.
        """
        pass

    def restore(self, commit_id):
        pass


class RepoRevision(object):
    def __init__(self, client, repo, commit_id):
        self.client = client
        self.repo = repo
        self.commit_id = commit_id

    def restore(self):
        """Restore the repo to this revision"""
        self.repo.revert(self.commit_id)



# coding: utf-8

import os
import pytest

import ebrains_drive
from tests.utils import randstring


USER = os.environ.get("SEAFILE_TEST_USERNAME", "test@seafiletest.com")
PASSWORD = os.environ.get("SEAFILE_TEST_PWD", None)
TOKEN = os.environ.get("SEAFILE_TEST_TOKEN", None)


@pytest.fixture(scope="session")
def client():
    if TOKEN:
        return ebrains_drive.client.DriveApiClient(username=USER, token=TOKEN, env="int")
    elif PASSWORD:
        return ebrains_drive.client.DriveApiClient(username=USER, password=PASSWORD, env="int")
    else:
        pytest.skip(
            "Must define one of the following environment variables: " "SEAFILE_TEST_PWD or SEAFILE_TEST_TOKEN"
        )


@pytest.fixture(scope="function")
def repo(client):
    repo_name = "tmp-测试资料库-%s" % randstring()
    repo = client.repos.create_repo(repo_name)
    try:
        yield repo
    finally:
        repo.delete()



import os
import string
import random


def randstring(length=12):
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def datafile(filename):
    return os.path.join(os.path.dirname(__file__), "data", filename)


def filesize(path):
    return os.stat(path).st_size



"""
A Python package for working with the Human Brain Project Model Validation Framework.

Andrew Davison and Shailesh Appukuttan, CNRS, 2017-2020

License: BSD 3-clause, see LICENSE.txt

"""

from ebrains_drive.client import DriveApiClient, BucketApiClient


def connect(username=None, password=None, token=None, env=""):
    client = DriveApiClient(username, password, token, env)
    return client



class ClientHttpError(Exception):
    """This exception is raised if the returned http response is not as
    expected"""

    def __init__(self, code, message):
        super(ClientHttpError, self).__init__()
        self.code = code
        self.message = message

    def __str__(self):
        return "ClientHttpError[%s: %s]" % (self.code, self.message)


class OperationError(Exception):
    """Expcetion to raise when an opeartion is failed"""

    pass


class DoesNotExist(Exception):
    """Raised when not matching resource can be found."""

    def __init__(self, msg):
        super(DoesNotExist, self).__init__()
        self.msg = msg

    def __str__(self):
        return "DoesNotExist: %s" % self.msg


class Unauthorized(Exception):

    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return "Unauthorized. This could be a result of either incorrect path or insufficient privilege. %s" % self.msg


class InvalidParameter(Exception):
    pass


class TokenExpired(Exception):
    pass


class UpstreamAPIException(Exception):
    """This exception is raised if the upstream API returns something unexpected"""



import os.path
import re

from ebrains_drive.repos import Repos
from ebrains_drive.files import SeafFile


class File(object):
    def __init__(self, client):
        self.client = client

    def get_file_by_url(self, file_url):
        """Get a single repo associated with specified repo_url
        Example inputs:
        1) https://drive.ebrains.eu/lib/0fee1620-062d-4643-865b-951de1eee355/file/sample-latest.csv
        2) https://drive.ebrains.eu/lib/0fee1620-062d-4643-865b-951de1eee355/file/Dir1/data.json
        """

        regex = r".*\/lib\/(.*)\/file(\/.*)"

        matches = re.search(regex, file_url)
        if matches is None:
            raise ValueError("Parameter `file_url` does not have expected format!")
        else:
            repo_id = matches.group(1)
            file_path = matches.group(2)

        repo_obj = self.client.repos.get_repo(repo_id)
        file_obj = repo_obj.get_file(file_path)
        return file_obj

    def get_file_by_local_path(self, local_path):
        """
        Get the file or directory object corresponding to `local_path`
        when the Drive is mounted in the EBRAINS Lab.
        """
        home_dir = "/mnt/user/drive/My Libraries/My Library/"
        group_dir = "/mnt/user/drive/Shared with groups/"
        shared_dir = "/mnt/user/shared/"
        if local_path.startswith(home_dir):
            repo_obj = self.client.repos.get_default_repo()
            relative_path = os.path.relpath(local_path, home_dir)
        elif local_path.startswith(group_dir):
            collab_name = local_path.split("/")[5]
            repo_obj = self.client.repos.get_repos_by_name(collab_name)
            relative_path = os.path.relpath(local_path, f"{group_dir}{collab_name}/")
        elif local_path.startswith(shared_dir):
            collab_name = local_path.split("/")[4]
            repo_obj = self.client.repos.get_repos_by_name(collab_name)
            relative_path = os.path.relpath(local_path, f"{shared_dir}{collab_name}/")
        else:
            raise Exception("Couldn't identify any file associated with specified path.")
        if len(repo_obj) == 0:
            raise Exception("Couldn't identify any file associated with specified path.")
        elif len(repo_obj) > 1:
            raise Exception("Couldn't uniquely identify the repo associated with specified path.")
        else:
            return repo_obj[0].get_file("/" + relative_path)



import inspect
import pytest
from ebrains_drive.utils import on_401_raise_unauthorized


def generator_fn():
    yield 1


test_401_parameters = [(generator_fn, True), (lambda: 1, False)]


@pytest.mark.parametrize("func,is_generator", test_401_parameters)
def test_on_401_wrap(func, is_generator):
    wrapped_fn = on_401_raise_unauthorized("oh noes")(func)
    assert inspect.isgeneratorfunction(wrapped_fn) == is_generator



import os
import re

from ebrains_drive.repo import Repo
from ebrains_drive.utils import raise_does_not_exist


class Repos(object):
    def __init__(self, client):
        self.client = client

    def create_repo(self, name, password=None):
        data = {"name": name}
        if password:
            data["passwd"] = password
        repo_json = self.client.post("/api2/repos/", data=data).json()
        return self.get_repo(repo_json["repo_id"])

    @raise_does_not_exist("The requested library does not exist")
    def get_repo(self, repo_id):
        """Get the repo which has the id `repo_id`.

        Raises :exc:`DoesNotExist` if no such repo exists.
        """
        repo_json = self.client.get("/api2/repos/" + repo_id).json()
        return Repo.from_json(self.client, repo_json)

    def _remove_duplicate_repos(self, repos):
        unique_repos = []
        for repo in repos:
            if repo.id not in [r.id for r in unique_repos]:
                unique_repos.append(repo)
            else:
                if repo.owner != "Organization":
                    unique_repos = [repo if r.id == repo.id else r for r in unique_repos]
        return unique_repos

    def list_repos(self):
        repos_json = self.client.get("/api2/repos/").json()
        repos = [Repo.from_json(self.client, j) for j in repos_json]
        return self._remove_duplicate_repos(repos)

    def get_repos_by_filter(self, filter_name, filter_value):
        """Get all repos which have `filter_name` = `filter_value`."""
        repos_json = self.client.get("/api2/repos/").json()
        print
        match_repos = []
        for j in repos_json:
            if filter_name in j.keys() and j[filter_name] == filter_value:
                match_repos.append(Repo.from_json(self.client, j))
        return self._remove_duplicate_repos(match_repos)

    def get_repos_by_name(self, repo_name):
        """Get all repos which have the name `repo_name`."""
        return self.get_repos_by_filter("name", repo_name)

    def get_repo_by_url(self, repo_url):
        """Get a single repo associated with specified repo_url
        Example inputs:
        1) https://wiki.ebrains.eu/bin/view/Collabs/collab-testing/subpage
        2) wiki.ebrains.eu/bin/view/Collabs/collab-testing
        3) collab-testing
        """

        regex = r"(?:\/Collabs\/)(.*?)(?:\/.*$|$)"

        matches = re.search(regex, repo_url)
        if matches is None:
            collab_name = repo_url
        else:
            collab_name = matches.group(1)

        match_repos = self.get_repos_by_filter("owner", "collab-" + collab_name + "-administrator")
        if not match_repos:
            match_repos = self.get_repos_by_filter("owner", "collab-" + collab_name + "-editor")
        if not match_repos:
            match_repos = self.get_repos_by_filter("owner", "collab-" + collab_name + "-viewer")

        if len(match_repos) == 0:
            raise Exception("Couldn't identify any repo associated with specified URL!")
        elif len(match_repos) > 1:
            raise Exception("Couldn't uniquely identify the repo associated with specified URL!")
        else:
            return match_repos[0]

    def get_default_repo(self):
        """
        Get the user's default repo (i.e. "My Library")
        """
        repos_json = self.client.get("/api2/default-repo").json()
        assert repos_json.get("exists"), f"Default repo does not exist."

        repo_id = repos_json.get("repo_id")
        assert repo_id, f"Expected repo_id to be populated, but wasn't"
        return self.get_repo(repo_id)

    def get_repo_by_local_path(self, local_path):
        """
        Get the repo that contains `local_path` when the Drive
        is mounted in the EBRAINS Lab.
        """
        home_dir = "/mnt/user/drive/My Libraries/My Library/"
        group_dir = "/mnt/user/drive/Shared with groups/"
        shared_dir = "/mnt/user/shared/"
        if local_path.startswith(home_dir):
            repo = self.get_default_repo()
        elif local_path.startswith(group_dir):
            collab_name = local_path.split("/")[5]
            repo = self.get_repos_by_name(collab_name)
        elif local_path.startswith(shared_dir):
            collab_name = local_path.split("/")[4]
            repo = self.get_repos_by_name(collab_name)
        else:
            raise Exception("Couldn't identify any repo associated with specified path.")
        return repo



import io
import os
import posixpath
import re
import time
from typing import Any, Dict
import requests
from tqdm import tqdm
from ebrains_drive.utils import querystr, on_401_raise_unauthorized

# Note: only files and dirs with contents is assigned an ID; else their ID is set to all zeros
ZERO_OBJ_ID = "0000000000000000000000000000000000000000"


class _SeafDirentBase(object):
    """Base class for :class:`SeafFile` and :class:`SeafDir`.

    It provides implementation of their common operations.
    """

    isdir = None

    def __init__(self, repo, path, object_id, obj_type, size=0):
        """
        :param:`path` the full path of this entry within its repo, like
        "/documents/example.md"

        :param:`size` The size of a file. It should be zero for a dir.
        """
        self.client = repo.client
        self.repo = repo
        self.path = path
        self.id = object_id
        self.type = obj_type
        self.size = size

    @property
    def name(self):
        return posixpath.basename(self.path)

    def list_revisions(self):
        pass

    def delete(self):
        suffix = "dir" if self.isdir else "file"
        url = "/api2/repos/%s/%s/" % (self.repo.id, suffix) + querystr(p=self.path)
        resp = self.client.delete(url)
        return resp

    def rename(self, newname):
        """Change file/folder name to newname"""
        suffix = "dir" if self.isdir else "file"
        url = "/api2/repos/%s/%s/" % (self.repo.id, suffix) + querystr(p=self.path, reloaddir="true")
        postdata = {"operation": "rename", "newname": newname}
        resp = self.client.post(url, data=postdata)
        succeeded = resp.status_code == 200
        if succeeded:
            if self.isdir:
                new_dirent = self.repo.get_dir(os.path.join(os.path.dirname(self.path), newname))
            else:
                new_dirent = self.repo.get_file(os.path.join(os.path.dirname(self.path), newname))
            for key in list(self.__dict__.keys()):
                self.__dict__[key] = new_dirent.__dict__[key]
        return succeeded

    def _copy_move_task(self, operation, dirent_type, dst_dir, dst_repo_id=None):
        url = "/api/v2.1/copy-move-task/"
        src_repo_id = self.repo.id
        src_parent_dir = os.path.dirname(self.path)
        src_dirent_name = os.path.basename(self.path)
        dst_repo_id = dst_repo_id
        dst_parent_dir = dst_dir
        operation = operation
        dirent_type = dirent_type
        postdata = {
            "src_repo_id": src_repo_id,
            "src_parent_dir": src_parent_dir,
            "src_dirent_name": src_dirent_name,
            "dst_repo_id": dst_repo_id,
            "dst_parent_dir": dst_parent_dir,
            "operation": operation,
            "dirent_type": dirent_type,
        }
        return self.client.post(url, data=postdata)

    def copyTo(self, dst_dir, dst_repo_id=None):
        """Copy file/folder to other directory (also to a different repo)"""
        if dst_repo_id is None:
            dst_repo_id = self.repo.id

        dirent_type = "dir" if self.isdir else "file"
        resp = self._copy_move_task("copy", dirent_type, dst_dir, dst_repo_id)
        return resp.status_code == 200

    def moveTo(self, dst_dir, dst_repo_id=None):
        """Move file/folder to other directory (also to a different repo)"""
        if dst_repo_id is None:
            dst_repo_id = self.repo.id

        dirent_type = "dir" if self.isdir else "file"
        resp = self._copy_move_task("move", dirent_type, dst_dir, dst_repo_id)
        succeeded = resp.status_code == 200
        if succeeded:
            new_repo = self.client.repos.get_repo(dst_repo_id)
            dst_path = os.path.join(dst_dir, os.path.basename(self.path))
            if self.isdir:
                new_dirent = new_repo.get_dir(dst_path)
            else:
                new_dirent = new_repo.get_file(dst_path)
            for key in list(self.__dict__.keys()):
                self.__dict__[key] = new_dirent.__dict__[key]
        return succeeded

    def get_share_link(self):
        dirent_type = "dir" if self.isdir else "file"
        url = f"/api2/repos/{self.repo.id}/{dirent_type}/shared-link/"
        resp = self.client.put(url, data={"p": self.path}, expected=(200, 201))
        succeeded = resp.status_code in (200, 201)
        if succeeded:
            return resp.headers["Location"]
        else:
            return None


class SeafDir(_SeafDirentBase):
    isdir = True

    def __init__(self, *args, **kwargs):
        super(SeafDir, self).__init__(*args, **kwargs)
        self.entries = None
        self.entries = kwargs.pop("entries", None)

    def ls(self, entity_type=None, force_refresh=True):
        """List the entries in this dir.

        Return a list of objects of class :class:`SeafFile` or :class:`SeafDir`.
        """
        if entity_type and entity_type not in ["file", "dir"]:
            raise ValueError("Invalid value for parameter `entity_type`; must be 'file' or 'dir'!")
        if self.entries is None or force_refresh:
            self.load_entries()

        if entity_type:
            return [x for x in self.entries if x.type == entity_type]
        else:
            return self.entries

    def share_to_user(self, email, permission):
        url = "/api2/repos/%s/dir/shared_items/" % self.repo.id + querystr(p=self.path)
        putdata = {"share_type": "user", "username": email, "permission": permission}
        resp = self.client.put(url, data=putdata)
        return resp.status_code == 200

    def create_empty_file(self, name):
        """Create a new empty file in this dir.
        Return a :class:`SeafFile` object of the newly created file.
        """
        # TODO: file name validation
        path = posixpath.join(self.path, name)
        url = "/api2/repos/%s/file/" % self.repo.id + querystr(p=path, reloaddir="true")
        postdata = {"operation": "create"}
        resp = self.client.post(url, data=postdata)
        self.id = resp.headers["oid"]
        self.load_entries(resp.json())
        return SeafFile(self.repo, path, ZERO_OBJ_ID, "file", 0)

    def check_exists(self, name, entity_type=None):
        """Check if an entity with specified name exists in current directory
        Note: seafile doesn't allow even a sub-directory and file,
              within the same directory, to have the same name
        """
        entity_list = self.ls(entity_type=entity_type, force_refresh=True)
        for e in entity_list:
            if e.name == name:
                return e
        return False

    def mkdir(self, name):
        """Create a new sub folder right under this dir.

        Return a :class:`SeafDir` object of the newly created sub folder.
        """
        # check if entity with same name already exists
        if self.check_exists(name):
            raise FileExistsError("File/directory with name = `{}` already exists in current directory!".format(name))

        path = posixpath.join(self.path, name)
        url = "/api2/repos/%s/dir/" % self.repo.id + querystr(p=path, reloaddir="true")
        postdata = {"operation": "mkdir"}
        resp = self.client.post(url, data=postdata)
        self.id = resp.headers["oid"]
        self.load_entries(resp.json())

        # fetch and return created directory object
        return SeafDir(self.repo, path, ZERO_OBJ_ID, "dir")

    def download(self, name=None):
        """Download the entire contents of a directory as a zip file

        :param:name The name of the downloaded zip file.
            If None, the name of the directory (or repo name in case of root directory) would be used.

        Returns a dict in following format:
        {'zipped': NUM, 'total': NUM, 'failed': NUM, 'failed_reason': '', 'canceled': NUM}
        """
        download_token = self._get_download_token()
        url = "/api/v2.1/query-zip-progress/?token=%s" % (download_token)
        wait = True
        while wait:
            resp = self.client.get(url).json()
            if resp["total"] == resp["zipped"] + resp["failed"] + resp["canceled"]:
                wait = False
            else:
                time.sleep(1)
        if resp["total"] != resp["zipped"]:
            raise Exception(resp["failed_reason"])
        url = "%s/seafhttp/zip/%s" % (self.client.server, download_token)
        zip_data = self.client.get(url).content
        if name:
            name = name if name.endswith(".zip") else name + ".zip"
        else:
            name = "%s.zip" % (self.repo.name) if (self.path == "/") else "%s.zip" % (self.path.split("/")[-1])
        with open(name, "wb") as f:
            f.write(zip_data)
        return resp

    def _get_download_token(self):
        if self.path == "/":
            parent_dir = "/"
            dirents = [item.name for item in self.ls()]
        else:
            parent_dir = "/".join(self.path.split("/")[0:-1]) or "/"
            dirents = self.path.split("/")[-1]
        url = "/api/v2.1/repos/%s/zip-task/" % (self.repo.id)
        data = {
            "parent_dir": parent_dir,
            "dirents": dirents,
        }
        resp = self.client.post(url, data=data).json()
        return resp["zip_token"]

    def upload(self, fileobj, filename):
        """Upload a file to this folder.

        :param:fileobj :class:`File` like object
        :param:filename The name of the file

        Return a :class:`SeafFile` object of the newly uploaded file.
        """
        if isinstance(fileobj, bytes):
            fileobj = io.BytesIO(fileobj)
        upload_url = self._get_upload_link()
        files = {
            "file": (filename, fileobj),
            "parent_dir": self.path,
        }
        self.client.post(upload_url, files=files)
        return self.repo.get_file(posixpath.join(self.path, filename))

    def upload_local_file(self, filepath, name=None, overwrite=False):
        """Upload a file to this folder.

        :param:filepath The path to the local file
        :param:name The name of this new file. If None, the name of the local file would be used.

        Return a :class:`SeafFile` object of the newly uploaded file.
        """
        name = name or os.path.basename(filepath)

        # check if entity with same name already exists
        entity_obj = self.check_exists(name)
        if entity_obj:
            if overwrite:
                a = entity_obj.delete()
            else:
                raise FileExistsError(
                    "File/directory with name = `{}` already exists in current directory!".format(name)
                )

        with open(filepath, "rb") as fp:
            return self.upload(fp, name)

    def _get_upload_link(self):
        url = "/api2/repos/%s/upload-link/?p=%s" % (self.repo.id, self.path)
        resp = self.client.get(url)
        return re.match(r'"(.*)"', resp.text).group(1)

    def get_uploadable_sharelink(self):
        """Generate a uploadable shared link to this dir.

        Return the url of this link.
        """
        pass

    def load_entries(self, dirents_json=None):
        if dirents_json is None:
            url = "/api2/repos/%s/dir/" % self.repo.id + querystr(p=self.path)
            dirents_json = self.client.get(url).json()

        self.entries = [self._load_dirent(entry_json) for entry_json in dirents_json]

    def _load_dirent(self, dirent_json):
        path = posixpath.join(self.path, dirent_json["name"])
        if dirent_json["type"] == "file":
            return SeafFile(self.repo, path, dirent_json["id"], dirent_json["type"], dirent_json["size"])
        else:
            return SeafDir(self.repo, path, dirent_json["id"], dirent_json["type"], 0)

    @property
    def num_entries(self):
        if self.entries is None:
            self.load_entries()
        return len(self.entries) if self.entries is not None else 0

    def __str__(self):
        return "SeafDir[repo=%s, path=%s, entries=%s]" % (self.repo.id[:6], self.path, self.num_entries)

    __repr__ = __str__


class SeafFile(_SeafDirentBase):
    isdir = False

    def update(self, fileobj):
        """Update the content of this file"""
        pass

    def __str__(self):
        return "SeafFile[repo=%s, path=%s, size=%s]" % (self.repo.id[:6], self.path, self.size)

    __repr__ = __str__

    def get_download_link(self):
        return self._get_download_link()

    def _get_download_link(self):
        url = "/api2/repos/%s/file/" % self.repo.id + querystr(p=self.path)
        resp = self.client.get(url)
        return re.match(r'"(.*)"', resp.text).group(1)

    def get_content(self):
        """Get the content of the file"""
        url = self._get_download_link()
        return self.client.get(url).content


class DataproxyFile:
    session = requests.Session()

    def __init__(
        self, client, bucket, hash: str, last_modified: str, bytes: int, name: str, content_type: str, **kwargs
    ) -> None:
        # TODO kwargs may contain keys: storage, data

        self.client = client
        self.bucket = bucket

        self.hash = hash
        self.last_modified = last_modified
        self.bytes = bytes
        self.name = name
        self.content_type = content_type

    def __str__(self):
        return "DataproxyFile[bucket=%s, path=%s, size=%s]" % (self.bucket.name, self.name, self.bytes)

    __repr__ = __str__

    def get_download_link(self):
        """n.b. this download link expires in the order of seconds"""
        resp = self.client.get(
            f"/v1/{self.bucket.target}/{self.bucket.dataproxy_entity_name}/{self.name}", params={"redirect": False}
        )
        return resp.json().get("url")

    def get_content(self, *, progress=False):
        url = self.get_download_link()
        # Auth header must **NOT** be attached to the download link obtained, or we will get 401

        if not progress:
            return DataproxyFile.session.get(url).content

        content = bytearray()
        resp = DataproxyFile.session.get(url, stream=True)
        with tqdm(total=int(resp.headers.get("content-length")), leave=True) as progress:
            for c in resp.iter_content(4096):
                content.extend(c)
                progress.update(4096)
        return bytes(content)

    @classmethod
    def from_json(cls, client, bucket, file_json: Dict[str, Any]):
        return cls(client, bucket, **file_json)

    @on_401_raise_unauthorized("Unauthorized")
    def delete(self):
        resp = self.client.delete(f"/v1/{self.bucket.target}/{self.bucket.dataproxy_entity_name}/{self.name}")
        json_resp = resp.json()
        if "failures" in json_resp:
            assert len(json_resp.get("failures")) == 0
        else:
            assert "has been removed" in json_resp["detail"]



import pytest
from unittest.mock import patch, mock_open
from unittest.mock import MagicMock
from ebrains_drive.bucket import Bucket
from ebrains_drive.exceptions import ClientHttpError, Unauthorized
from io import StringIO, IOBase
from itertools import product


class MockClient:
    def get(self, *args, **kwargs):
        raise NotImplementedError

    def put(self, *args, **kwargs):
        raise NotImplementedError


@pytest.fixture
def mock_client():
    client = MockClient()
    client.get = MagicMock()
    client.put = MagicMock()
    return client


class MockHttpResp:
    def __init__(self, resp):
        self.resp = resp

    def json(self):
        return self.resp

    def raise_for_status(self): ...


bucket_json = {
    "name": "foo",
    "objects_count": 12,
    "bytes": 112233,
    "is_public": False,
    "role": "admin",
}

file_json1 = {
    "name": "foo",
    "hash": "hash-foo",
    "last_modified": "last-modified",
    "bytes": 123,
    "content_type": "json",
}


def test_from_json():
    client = MockClient()
    bucket = Bucket.from_json(client, bucket_json)
    assert isinstance(bucket, Bucket)


def test_ls_when_raise_client_error():

    client = MockClient()
    client.get = MagicMock()
    client.get.side_effect = [ClientHttpError(401, "foo-bar")]

    bucket = Bucket.from_json(client, bucket_json)

    try:
        fs = [f for f in bucket.ls()]
        raise Exception("did not raise")
    except Exception as e:
        assert isinstance(e, Unauthorized), f"Expect raise Unauthorized: {e}"


def test_ls_when_repeats():

    client = MockClient()
    client.get = MagicMock()
    client.get.side_effect = [MockHttpResp({"objects": [file_json1, file_json1]})]
    bucket = Bucket.from_json(client, bucket_json)

    try:
        fs = [f for f in bucket.ls()]
        raise Exception("did not raise")
    except Exception as e:
        assert isinstance(e, RuntimeError), f"Expect raise RuntimeError: {e}"


@pytest.fixture
def mocked_request():
    try:
        with patch("requests.request") as patched_obj:
            yield patched_obj
    finally:
        ...


@pytest.fixture
def mock_open_fixture():
    try:
        with patch("builtins.open", new_callable=mock_open, read_data="foo-bar") as patched_obj:
            yield patched_obj
    finally:
        ...


@pytest.mark.parametrize("filelike,kwargs", product(["filelike", StringIO()], [{"foo": "bar"}, {}]))
def test_upload(filelike, kwargs, mocked_request, mock_open_fixture, mock_client: MockClient):
    bucket = Bucket.from_json(mock_client, bucket_json)
    mocked_request.return_value = MockHttpResp({})
    mock_client.put.return_value = MockHttpResp({"url": "http://foo-bar.co/"})

    bucket.upload(filelike, "filename", **kwargs)
    if isinstance(filelike, str):
        mock_open_fixture.assert_called()
        data = mock_open_fixture.return_value
    elif isinstance(filelike, IOBase):
        mock_open_fixture.assert_not_called()
        data = filelike
    else:
        raise RuntimeError(f" should be either str or IOBase")
    mocked_request.assert_called_with("PUT", "http://foo-bar.co/", data=data, **kwargs)



from ebrains_drive.exceptions import ClientHttpError, Unauthorized
from ebrains_drive.utils import on_401_raise_unauthorized
from ebrains_drive.bucket import Bucket
from time import sleep


class Buckets(object):

    def __init__(self, client):
        self.client = client

    @on_401_raise_unauthorized(
        "401 response. Check you/your token have access right and/or the bucket name has been spelt correctly."
    )
    def get_bucket(self, bucket_name: str, *, public: bool = False) -> Bucket:
        """Get the specified bucket according name. If forced flag is set to True, will attempt to create the collab, if necessary."""
        resp = self.client.get(f"/v1/buckets/{bucket_name}/stat")
        return Bucket.from_json(self.client, resp.json(), public=public, target="buckets")

    def get_dataset(self, dataset_id: str, *, public: bool = False, request_access: bool = False):
        request_sent = False
        attempt_no = 0
        while True:
            try:
                resp = self.client.get(f"/v1/datasets/{dataset_id}/stat")
                return Bucket.from_json(
                    self.client, resp.json(), public=public, target="datasets", dataset_id=dataset_id
                )
            except ClientHttpError as e:
                if e.code != 401:
                    raise e

                if not request_access:
                    raise Unauthorized(
                        f"You do not have access to this dataset. If this is a private dataset, try to set request_access flag to true. We can start the procedure of requesting access for you."
                    )
                if not request_sent:
                    self.client.post(f"/v1/datasets/{dataset_id}", expected=(200, 201))
                    request_sent = True
                    print("Request sent. Please check the mail box associated with the token.")
                sleep(5)
                attempt_no = attempt_no + 1
                print(f"Checking permission, attempt {attempt_no}")



from getpass import getpass
import requests
from abc import ABC
import base64
import json
import time
from copy import copy, deepcopy
from ebrains_drive.utils import on_401_raise_unauthorized
from ebrains_drive.exceptions import ClientHttpError, TokenExpired, Unauthorized
from ebrains_drive.repos import Repos
from ebrains_drive.buckets import Buckets
from ebrains_drive.file import File


class ClientBase(ABC):
    def __init__(self, username=None, password=None, token=None, env="") -> None:

        self.username = username
        self.password = password
        self._token = token
        self.server = None
        self.session = requests.Session()

        if token is None:
            if self.username is None:
                self.username = input("EBRAINS username: ")
            if self.password is None:
                self.password = getpass()

            try:
                self._get_token()
            except KeyError:
                print("Error: Invalid user credentials!")
                raise

    def _set_env(self, env=""):
        self.suffix = ""

        if env == "dev":
            self.suffix = "-dev"
        elif env == "int":
            self.suffix = "-int"
        # else we keep empty suffix for production

        self.iam_host = "iam" + self.suffix + ".ebrains.eu"
        self.iam_url = "https://" + self.iam_host

    def _get_token(self):
        response = requests.post(
            self.iam_url + "/auth/realms/hbp/protocol/openid-connect/token",
            auth=("ebrains-drive", ""),
            data={"grant_type": "password", "username": self.username, "password": self.password, "scope": "openid"},
        )

        if response.status_code == 200:
            self._token = response.json()["access_token"]
        elif response.status_code == 401:
            raise Unauthorized(response.json()["error_description"])
        else:
            raise ClientHttpError(response.json()["error_description"])

    def get(self, *args, **kwargs):
        return self.send_request("GET", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.send_request("POST", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.send_request("PUT", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.send_request("DELETE", *args, **kwargs)

    def send_request(self, method: str, url: str, *args, **kwargs):
        if not url.startswith("http"):
            # sanity checks.
            # - accounts for if server was provided with trailing slashes
            # - accounts for if url was provided with leading slashes
            url = self.server.rstrip("/") + "/" + url.lstrip("/")

        # Copy the kwargs, and deepcopy the ones we change, so as not to mutate the original kwargs.
        # We cannot deepcopy the whole thing, because some values (e.g. BufferedReader objects)
        # cannot be pickled
        kwargs = copy(kwargs)
        headers = deepcopy(kwargs.get("headers", {}))
        headers.setdefault("Authorization", "Bearer " + self._token)
        kwargs["headers"] = headers

        expected = kwargs.pop("expected", 200)
        if not hasattr(expected, "__iter__"):
            expected = (expected,)
        resp = self.session.request(method, url, *args, **kwargs)
        if resp.status_code not in expected:
            msg = "Expected %s, but get %s" % (" or ".join(map(str, expected)), resp.status_code)
            raise ClientHttpError(resp.status_code, msg)

        return resp


class DriveApiClient(ClientBase):
    """Wraps seafile web api"""

    def __init__(self, username=None, password=None, token=None, env=""):
        """Wraps various basic operations to interact with seahub http api."""
        self._set_env(env)
        super().__init__(username, password, token, env)

        self.server = self.drive_url

        self.repos = Repos(self)
        self.groups = Groups(self)
        self.file = File(self)

    def _set_env(self, env=""):
        super()._set_env(env)
        self.drive_url = "https://drive" + self.suffix + ".ebrains.eu"

    def get_drive_url(self):
        return self.drive_url

    def get_iam_host(self):
        return self.iam_host

    def get_iam_url(self):
        return self.iam_url

    def __str__(self):
        return "DriveApiClient[server=%s, user=%s]" % (self.server, self.username)

    __repr__ = __str__

    def send_request(self, method: str, url: str, *args, **kwargs):
        if not url.startswith("http"):
            assert not self.server.endswith("/")
            if url.startswith("/"):
                url = f"{self.server}{url}"
            else:
                url = f"{self.server}/{url}"
        return super().send_request(method, url, *args, **kwargs)


_I_AM_A_PUBLIC_BUCKET = "_I_AM_A_PUBLIC_BUCKET"


class BucketApiClient(ClientBase):

    def __init__(self, username=None, password=None, token=_I_AM_A_PUBLIC_BUCKET, env="") -> None:
        self._set_env(env)

        super().__init__(username, password, token, env)

        self.server = f"https://data-proxy{self.suffix}.ebrains.eu/api"

        self.buckets = Buckets(self)

    @on_401_raise_unauthorized(
        "Failed. Note: BucketApiClient.create_new needs to have clb.drive:write as a part of scope."
    )
    def create_new(self, bucket_name: str, title=None, description="Created by ebrains_drive"):
        """
        Create a new bucket by first attempting to create a new wiki/collab. On 201 (created)
        or 409 (conflict) initialize the bucket of the said wiki. The request to initialize the bucket
        will be retried up to 5 times, as it usually takes a few minutes for the newly initialized wiki
        to allow buckets to be created.

        :param:`bucket_name` the name of the to-be-created bucket (and wiki if needed)

        :param:`title` the title of the to-be-created wiki (if unset, defaults to `bucket_name` param)

        :param:`description` description of the to-be-created wiki.
        """

        self.send_request(
            "POST",
            f"https://wiki{self.suffix}.ebrains.eu/rest/v1/collabs",
            json={
                "name": bucket_name,
                "title": title or bucket_name,
                "description": description,
                "drive": True,
                "chat": True,
                "public": False,
            },
            expected=(201, 409),
        )

        fuse = 5
        while True:
            try:
                self.send_request("POST", "/v1/buckets", json={"bucket_name": bucket_name}, expected=201)
                break
            except Exception as e:
                if fuse < 0:
                    raise e from e
                fuse -= 1
                time.sleep(1)

    @on_401_raise_unauthorized(
        "Failed. Note: BucketApiClient.delete_bucket needs to have clb.drive:write as a part of scope."
    )
    def delete_bucket(self, bucket_name: str, *, delete_wiki=False):
        """
        Deletes an existing bucket.

        :param:`bucket_name` name of the bucket (and - if delete_wiki is set - of the wiki) to be deleted

        :param:`delete_wiki` if the wiki should also be deleted.
        """
        self.send_request("DELETE", f"/v1/buckets/{bucket_name}", expected=(200,))
        if delete_wiki:
            self.send_request("DELETE", f"https://wiki.ebrains.eu/rest/v1/collabs/{bucket_name}", expected=(200,))

    def send_request(self, method: str, url: str, *args, **kwargs):

        if self._token != _I_AM_A_PUBLIC_BUCKET:
            hdr, info, sig = self._token.split(".")
            info_json = base64.b64decode(info + "==").decode("utf-8")

            # https://www.rfc-editor.org/rfc/rfc7519#section-2
            exp_utc_seconds = json.loads(info_json).get("exp")
            now_tc_seconds = time.time()

            if now_tc_seconds > exp_utc_seconds:
                raise TokenExpired

        if self._token == _I_AM_A_PUBLIC_BUCKET:
            headers = kwargs.get("headers", {})
            headers["Authorization"] = None
            kwargs["headers"] = headers

        return super().send_request(method, url, *args, **kwargs)


class Groups(object):
    def __init__(self, client):
        pass

    def create_group(self, name):
        pass



import string
import random
import inspect
from functools import wraps
from typing import Type
from urllib.parse import urlencode
from ebrains_drive.exceptions import ClientHttpError, DoesNotExist, Unauthorized


def randstring(length=0):
    if length == 0:
        length = random.randint(1, 30)
    return "".join(random.choice(string.lowercase) for i in range(length))


def _raise_on(http_code: int, Ex: Type[Exception]):
    """Decorator factory funciton to turn a function that get a http http_code response
    to a `Ex` exception."""

    def raise_on(msg: str):
        def decorator(func):

            if inspect.isgeneratorfunction(func):

                @wraps(func)
                def wrapped(*args, **kwargs):
                    try:
                        yield from func(*args, **kwargs)
                    except ClientHttpError as e:
                        if e.code == http_code:
                            raise Ex(msg)
                        else:
                            raise e

                return wrapped

            else:

                @wraps(func)
                def wrapped(*args, **kwargs):
                    try:
                        return func(*args, **kwargs)
                    except ClientHttpError as e:
                        if e.code == http_code:
                            raise Ex(msg)
                        else:
                            raise e

                return wrapped

        return decorator

    return raise_on


on_401_raise_unauthorized = _raise_on(401, Unauthorized)


def raise_does_not_exist(msg):
    """Decorator to turn a function that get a http 404 response to a
    :exc:`DoesNotExist` exception."""

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ClientHttpError as e:
                if e.code == 404:
                    raise DoesNotExist(msg)
                else:
                    raise

        return wrapped

    return decorator


def to_utf8(obj):
    if isinstance(obj, str):
        return obj.encode("utf-8")
    return obj


def querystr(**kwargs):
    return "?" + urlencode(kwargs)


# not used?
def utf8lize(obj):
    if isinstance(obj, dict):
        return {k: to_utf8(v) for k, v in obj.items()}

    if isinstance(obj, list):
        return [to_utf8(x) for x in ob]

    if instance(obj, str):
        return obj.encode("utf-8")

    return obj



import os

os.environ["EBRAINS_AUTH_ENDPOINT"] = "https://iam-int.ebrains.eu/auth/realms/hbp"

from ebrains_iam.client_credential import ClientCredentialsSession

sess = ClientCredentialsSession(
    client_id=os.environ["EBRAINS_CLIENT_ID"], client_secret=os.environ["EBRAINS_CLIENT_SECRET"]
)
token = sess.get_token()

print(token)



from typing import Iterable
import requests
from ebrains_drive.exceptions import DoesNotExist, InvalidParameter, UpstreamAPIException
from ebrains_drive.files import DataproxyFile
from ebrains_drive.utils import on_401_raise_unauthorized
from io import IOBase
from typing import Union


class Bucket(object):

    LIMIT = 100

    """
    A dataproxy bucket
    n.b. for a dataset bucket, role & is_public may be None
    """

    def __init__(
        self,
        client,
        name: str,
        objects_count: int,
        bytes: int,
        last_modified: str = None,
        is_public: bool = None,
        is_initialized: bool = None,
        role: str = None,
        *,
        public: bool = False,
        target: str = "buckets",
        dataset_id: str = None,
    ) -> None:
        if target != "buckets" and target != "datasets":
            raise InvalidParameter(
                f"Init Buckets exception: target can be left unset, but if set, must either be buckets or datasets"
            )
        if public:
            raise NotImplementedError(f"Access to public datasets/buckets NYI.")
        self.public = public
        self.target = target

        self.client = client

        self.name = name
        self.objects_count = objects_count
        self.bytes = bytes
        self.last_modified = last_modified
        self.is_public = is_public
        self.role = role
        self.is_initialized = is_initialized

        # n.b. for dataset bucket, dataset_id needs to be used for dataproxy_entity_name, but for collab bucket, name is used
        self.dataproxy_entity_name = dataset_id or name

    @classmethod
    def from_json(
        cls, client, bucket_json, *, public: bool = False, target: str = "buckets", dataset_id=None
    ) -> "Bucket":
        return cls(client, **bucket_json, public=public, target=target, dataset_id=dataset_id)

    def __str__(self):
        return "(name='{}')".format(self.name)

    def __repr__(self):
        return "ebrains_drive.bucket.Bucket(name='{}')".format(self.name)

    @on_401_raise_unauthorized("Unauthorized.")
    def ls(self, prefix: str = None) -> Iterable[DataproxyFile]:
        marker = None
        visited_name = set()
        while True:
            resp = self.client.get(
                f"/v1/{self.target}/{self.dataproxy_entity_name}",
                params={"limit": self.LIMIT, "marker": marker, "prefix": prefix},
            )
            objects = resp.json().get("objects", [])
            if len(objects) == 0:
                break

            for obj in objects:

                yield DataproxyFile.from_json(self.client, self, obj)
                marker = obj.get("name")

                if marker in visited_name:
                    raise RuntimeError(f"Bucket.ls error: hash {marker} has already been visited.")
                visited_name.add(marker)
        return

    @on_401_raise_unauthorized("Unauthorized")
    def get_file(self, name: str) -> DataproxyFile:
        name = name.lstrip("/")
        for file in self.ls(prefix=name):
            if file.name == name:
                return file
        raise DoesNotExist(f"Cannot find {name}.")

    @on_401_raise_unauthorized("Unauthorized")
    def upload(self, filelike: Union[str, IOBase], filename: str, **kwargs):
        filename = filename.lstrip("/")
        resp = self.client.put(f"/v1/{self.target}/{self.dataproxy_entity_name}/{filename}", **kwargs)
        upload_url = resp.json().get("url")
        if upload_url is None:
            raise UpstreamAPIException(f"Bucket.upload did not get upload url.")
        filehandle = filelike if isinstance(filelike, IOBase) else open(filelike, "rb")
        resp = requests.request("PUT", upload_url, data=filehandle, **kwargs)
        resp.raise_for_status()
